﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 * and the BEACON Center for the Study of Evolution in Action.
 * 
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  [Item("RegressionTreeModel", "Represents a decision tree for regression.")]
  public class RegressionTreeModel : NamedItem, IRegressionModel {

    // trees are represented as a flat array
    public struct TreeNode {
      public readonly static string NO_VARIABLE = string.Empty;
      public string varName; // name of the variable for splitting or -1 if terminal node
      public double val; // threshold
      public int leftIdx;
      public int rightIdx;

      public override int GetHashCode() {
        return leftIdx ^ rightIdx ^ val.GetHashCode();
      }
    }

    [Storable]
    public readonly TreeNode[] tree;

    [StorableConstructor]
    private RegressionTreeModel(bool serializing) : base(serializing) { }
    // cloning ctor
    public RegressionTreeModel(RegressionTreeModel original, Cloner cloner)
      : base(original, cloner) {
      this.tree = original.tree; // shallow clone, tree must be readonly
    }

    public RegressionTreeModel(TreeNode[] tree)
      : base("RegressionTreeModel", "Represents a decision tree for regression.") {
      this.tree = tree;
    }

    private static double GetPredictionForRow(TreeNode[] t, int nodeIdx, IDataset ds, int row) {
      var node = t[nodeIdx];
      if (node.varName == TreeNode.NO_VARIABLE)
        return node.val;
      else if (ds.GetDoubleValue(node.varName, row) <= node.val)
        return GetPredictionForRow(t, node.leftIdx, ds, row);
      else
        return GetPredictionForRow(t, node.rightIdx, ds, row);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RegressionTreeModel(this, cloner);
    }

    public IEnumerable<double> GetEstimatedValues(IDataset ds, IEnumerable<int> rows) {
      return rows.Select(r => GetPredictionForRow(tree, 0, ds, r));
    }

    public IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, new RegressionProblemData(problemData));
    }
  }

}
